/*
	CASIO Z-1/FX-890P emulator
*/

#ifndef Z1_H
#define Z1_H

#include <stdio.h>
#include <limits.h>
#include "i80x86.h"
#include "conf.h"

/* ^U */
#define FALSE	0	/* U */
#define TRUE	1	/* ^ */

/* @ */
#define MACHINE_Z1	0	/* Z-1 */
#define MACHINE_Z1GR	1	/* Z-1GR */
#define MACHINE_Z1GRA	2	/* Z-1GRa */
#define MACHINE_FX890P	3	/* FX-890P */
#define MACHINE_FX890P_EN	4	/* FX-890PCOf */

/* LCD */
#define VRAM_HEIGHT	65	/* VRAM  */
#define VRAM_WIDTH	32	/* VRAM oCg */
#define LCD_HEIGHT	32	/* LCD  */
#define LCD_WIDTH	192	/* LCD hbg */

/* L[R[h */
#define ZKEY_BRK	0x01	/* BRK/ONL[ */
#define ZKEY_TAB	0x12	/* TABL[ */
#define ZKEY_Q	0x13	/* QL[ */
#define ZKEY_ALLRESET	0x14	/* I[Zbg{^ */
#define ZKEY_A	0x15	/* AL[ */
#define ZKEY_CAPS	0x16	/* CAPS/JiL[ */
#define ZKEY_Z	0x17	/* ZL[ */
#define ZKEY_W	0x22	/* WL[ */
#define ZKEY_E	0x23	/* EL[ */
#define ZKEY_S	0x24	/* SL[ */
#define ZKEY_D	0x25	/* DL[ */
#define ZKEY_X	0x26	/* XL[ */
#define ZKEY_C	0x27	/* XL[ */
#define ZKEY_SRCH	0x28	/* SRCHL[ */
#define ZKEY_IN	0x29	/* INL[ */
#define ZKEY_R	0x32	/* RL[ */
#define ZKEY_T	0x33	/* TL[ */
#define ZKEY_F	0x34	/* FL[ */
#define ZKEY_G	0x35	/* GL[ */
#define ZKEY_V	0x36	/* VL[ */
#define ZKEY_B	0x37	/* BL[ */
#define ZKEY_OUT	0x38	/* OUTL[ */
#define ZKEY_CALC	0x39	/* CALCL[ */
#define ZKEY_Y	0x42	/* YL[ */
#define ZKEY_U	0x43	/* UL[ */
#define ZKEY_H	0x44	/* HL[ */
#define ZKEY_J	0x45	/* JL[ */
#define ZKEY_N	0x46	/* NL[ */
#define ZKEY_M	0x47	/* ML[ */
#define ZKEY_SPC	0x48	/* Xy[XL[ */
#define ZKEY_EQUAL	0x49	/* =L[ */
#define ZKEY_I	0x52	/* IL[ */
#define ZKEY_O	0x53	/* OL[ */
#define ZKEY_K	0x54	/* KL[ */
#define ZKEY_L	0x55	/* LL[ */
#define ZKEY_COMMA	0x56	/* ,L[ */
#define ZKEY_INS	0x57	/* INSL[ */
#define ZKEY_LEFT	0x58	/* L[ */
#define ZKEY_DOWN	0x59	/* L[ */
#define ZKEY_P	0x62	/* PL[ */
#define ZKEY_2ND	0x63	/* [S]L[ */
#define ZKEY_SEMICOLON	0x64	/* ;L[ */
#define ZKEY_COLON	0x65	/* :L[ */
#define ZKEY_UP	0x66	/* L[ */
#define ZKEY_DEL	0x67	/* DELL[ */
#define ZKEY_RIGHT	0x68	/* L[ */
#define ZKEY_0	0x69	/* 0L[ */
#define ZKEY_MENU	0x72	/* MENUL[ */
#define ZKEY_LOG	0x73	/* LOGL[ */
#define ZKEY_MR	0x74	/* MRL[ */
#define ZKEY_7	0x75	/* 7L[ */
#define ZKEY_4	0x76	/* 4L[ */
#define ZKEY_1	0x77	/* 1L[ */
#define ZKEY_PERIOD	0x78	/* .L[ */
#define ZKEY_EXP	0x79	/* EL[ */
#define ZKEY_RETURN	0x7a	/* ^[L[ */
#define ZKEY_CAL	0x82	/* CALL[ */
#define ZKEY_LN	0x83	/* lnL[ */
#define ZKEY_MPLUS	0x84	/* M+L[ */
#define ZKEY_8	0x85	/* 8L[ */
#define ZKEY_5	0x86	/* 5L[ */
#define ZKEY_2	0x87	/* 2L[ */
#define ZKEY_3	0x88	/* 3L[ */
#define ZKEY_PLUS	0x89	/* +L[ */
#define ZKEY_MINUS	0x8a	/* -L[ */
#define ZKEY_DEGR	0x92	/* DEGRL[ */
#define ZKEY_SIN	0x93	/* sinL[ */
#define ZKEY_LPAREN	0x94	/* (L[ */
#define ZKEY_RPAREN	0x95	/* )L[ */
#define ZKEY_9	0x96	/* 9L[ */
#define ZKEY_6	0x97	/* 6L[ */
#define ZKEY_ASTER	0x98	/* *L[ */
#define ZKEY_SLASH	0x99	/* /L[ */
#define ZKEY_BS	0x9a	/* BSL[ */
#define ZKEY_SQR	0xa2	/* L[ */
#define ZKEY_X2	0xa3	/* x^2L[ */
#define ZKEY_ENG	0xa4	/* ENGL[ */
#define ZKEY_CLS	0xa5	/* CLSL[ */
#define ZKEY_COS	0xa6	/* cosL[ */
#define ZKEY_HAT	0xa7	/* ^L[ */
#define ZKEY_ANS	0xa8	/* ANSL[ */
#define ZKEY_TAN	0xaa	/* tanL[ */
#define ZKEY_SHIFT	0xbc	/* SHIFTL[ */
#define ZKEY_COPY	0xfa	/* Rs[zL[ */
#define ZKEY_PASTE	0xfb	/* \tzL[ */
#define ZKEY_REWIND_INPORT	0xfc	/* ͊߂zL[ */
#define ZKEY_REWIND_OUTPORT	0xfd	/* o͊߂zL[ */
#define ZKEY_DEBUG	0xfe	/* fobOzL[ */
#define ZKEY_OFF	0xff	/* dOFFzL[ */

/* Rrl[VL[ */
#define ZKEYMOD_NOSHIFT	0x4000	/* SHIFTȂ */
#define ZKEYMOD_SHIFT	0x8000	/* SHIFT */
#define ZKEYMOD_MASK	0xc000	/* Rrl[VL[}XN */

/* TEh[g(Hz) */
#define AUDIO_RATE	44100

/* 傫ق߂ */
#define MAX(x, y)	((x) > (y) ? (x): (y))

/* ق߂ */
#define MIN(x, y)	((x) < (y) ? (x): (y))

/* ʁEʃoCg */
#if defined(I86_LITTLEENDIAN)
#	define LOW(x)	((uint8 *)&(x) + 0)
#	define HIGH(x)	((uint8 *)&(x) + 1)
#else
#	define LOW(x)	((uint8 *)&(x) + 1)
#	define HIGH(x)	((uint8 *)&(x) + 0)
#endif

/* 64rbgȂ */
typedef unsigned long long	uint64;
/* 64rbg */
typedef long long	int64;

/*
	^C}
*/
struct Timer {
		uint16 interval_a; /* ԊuA */
		uint16 interval_b; /* ԊuB */
		uint16 control; /*  */
		uint16 count; /* JE^ */
};

/*
	o̓|[g
*/
struct IOPort {
	char path[PATH_MAX]; /* o̓t@C̃pX */
	int pos; /* Ǎʒu */
};

/*
	Z-1/FX-890P̏
*/
typedef struct {
	I86stat cpu; /* CPU */
	uint8 memory[0x40000 + 0x20000]; /*  */

	uint16 power; /* d */
	uint8 buzzer; /* uU[ */

	/* ^C} */
	struct {
		struct Timer t0; /* ^C}0 */
		struct Timer t1; /* ^C}1 */
		struct Timer t2; /* ^C}2 */
		uint16 control; /* Rg[ */
		int intr; /* ? */
	} timer;

	/* VA|[g */
	struct {
		uint16 baud; /* {[[g */
		uint16 settings; /* ݒ */
		uint16 control; /* Rg[ */
		int intr; /* ? */
		int sent; /* Mς? */
	} sio;

	/* J[hGbW */
	struct {
		uint16 control; /* Rg[ */
		int intr; /* ? */
	} card;

	/* L[ */
	struct {
		uint16 strobe; /* Xg[u */
		uint16 matrix[12]; /* }gNX */
		uint16 key_control; /* L[Rg[ */
		int key_intr; /* L[? */
		uint16 control; /* Rg[ */
		int intr; /* ? */
	} key;

	/* dXCb` */
	struct {
		uint16 control; /* Rg[ */
		int intr; /* ? */
	} sw;

	/* FDD */
	struct {
		uint8 cmd; /* R}h */
		uint8 param[32], *param_p; /* p[^ */
		uint8 err; /* G[ */
	} fdd;

	/* VRAM */
	struct {
		uint8 ar; /* \ */
		uint8 xar; /* XAhXWX^ */
		uint8 yar; /* YAhXWX^ */
		uint8 fcr; /* Rg[WX^ */
		uint8 mdr; /* [hWX^ */
		uint8 csr; /* CZNgWX^ */
		uint8 vram[(LCD_WIDTH / 8) * LCD_HEIGHT]; /* LCD */
		uint8 status; /* V{KCh(01:CAPS, 04:S, 08:BASIC, 10:DEG, 20:RAD, 80:GRA) */
	} vram;

	/* Z */
	struct {
		uint8 op; /* Z */
		uint8 map; /* Zʂ}bv邩? */
		uint8 reg[9]; /* WX^ */
	} cal;

	/* LCD */
	struct Lcd {
		int pix[LCD_HEIGHT * LCD_WIDTH]; /* hbg̃sNZ̔Zx */
		int symbol[6]; /* V{̔Zx */
	} lcd;

	/* VAMf[^ */
	struct IOPort rs_receive;

	/* VAMf[^ */
	struct IOPort rs_send;

	/* v^o̓f[^ */
	struct IOPort printer;

	/* tbs[fBXN */
	struct Disk {
		char dir[PATH_MAX]; /* ztbs[fBXÑfBNg */
		FILE *fp; /* I[vt@C̃fBXNv^ */
		size_t size; /* I[vt@C̃TCY */
		off_t pos; /* I[vt@C̓Ǎ/ʒu */
		int result_count; /* ʓǍJE^ */
		uint16 len; /* obt@f[^ */
		uint8 *files; /* t@C */
		uint8 *files_p; /* t@Cʂ̓ǎ|C^ */
		uint8 data; /* Ǎf[^ */
	} disk;

	/* TEh */
	struct Sound {
		int8 *buffer[2]; /* TEhobt@ */
		int len; /* obt@ */
		int page; /* Đ̃y[W */
		int last_states; /* Ōɏo͂_̌o߃Xe[g */
		int8 last_vol; /* Ōɏ񂾃{[ */
		int last_pos; /* Ōɏ񂾈ʒu */
		int played; /* Đ? */
	} sound;

	/* WCXeBbN */
	struct Joystick {
		uint8 right;	/* EɊ蓖ĂꂽL[ */
		uint8 left;	/* Ɋ蓖ĂꂽL[ */
		uint8 up;	/* Ɋ蓖ĂꂽL[ */
		uint8 down;	/* Ɋ蓖ĂꂽL[ */
		uint8 button[16];	/* {^Ɋ蓖ĂꂽL[ */
	} joy;

	/*  */
	struct AutoKey {
		uint8 *key; /* ͒̃L[ */
		uint8 press; /* Ă邩? */
		int count; /* ҂JE^ */
	} auto_key;

	/* ݒ */
	struct Setting {
		int machine; /* @ */
		const char *machine_name; /* @햼 */
		int cpu_clock; /* CPUNbN(Hz) */
		uint16 iram_size;	/* RAMe */
		uint16 oram_size;	/* gRAMe */
		int refresh_rate; /* I/OXV[g(Hz) */
		int zoom; /* LCD{ */
		int scales;	/* LCDK */
		char path_rom[PATH_MAX]; /* ROMC[Wt@C̃pX */
		char path_ram[PATH_MAX]; /* RAMC[Wt@C̃pX */
	} setting;
} Z1stat;

/* init.c */
int getAudioSamples(int *fps);
int init(Z1stat *, int, char **);

/* depend.c */
void showError(const char *, ...);
void updateWindow(Z1stat *);
int updateKey(Z1stat *);
void delay(int);
void setTitle(const char *);
int initDepend(Z1stat *);

/* z1.c */
char *getDirName(char *, const char *);
int getTimerCount(Z1stat *, const struct Timer *);
void setTimerCount(Z1stat *, struct Timer *, int);
int sendIOData(struct IOPort *, uint8);
int peekIOData(struct IOPort *, uint8 *);
int receiveIOData(struct IOPort *, uint8 *);
void setIOData(struct IOPort *, const char *);
int64 decodeMan(const uint8 *bcd);
uint8 *encodeMan(uint8 *bcd, int64 man);
void opXY(uint8, const uint8 *, const uint8 *, uint8 *, uint8 *);
char *decodeValue(char *, const uint8 *);
void writeSound(Z1stat *, uint8);
void flipSoundBuffer(Z1stat *);
uint8 getFdFreeSize(struct Disk *, uint32 *);
uint8 openFdFile(struct Disk *, const uint8 *, const uint8 *);
uint8 closeFdFile(struct Disk *);
uint8 readFdFile(struct Disk *, uint8 *);
uint8 writeFdFile(struct Disk *, uint8, const uint8 *);
uint8 renameFdFile(struct Disk *, const uint8 *, const uint8 *);
uint8 deleteFdFile(struct Disk *, const uint8 *);
uint8 findFdFile(struct Disk *, const uint8 *, uint8 *);
uint8 getFoundFdFile(struct Disk *, uint8 *);
uint8 formatFd(struct Disk *);
int autoUpdateKey(Z1stat *);
void setAutoKey(Z1stat *, uint8);
void setAutoText(Z1stat *, const char *);
int pressKey(Z1stat *, uint8);
int releaseKey(Z1stat *, uint8);

#if defined(DEF_GLOBAL)
#	define EXTERN
#else
#	define EXTERN	extern
#endif

#include "depend.h"

#endif

/*
	Copyright 2009~2019 maruhiro
	All rights reserved. 

	Redistribution and use in source and binary forms, 
	with or without modification, are permitted provided that 
	the following conditions are met: 

	 1. Redistributions of source code must retain the above copyright notice, 
	    this list of conditions and the following disclaimer. 

	 2. Redistributions in binary form must reproduce the above copyright notice, 
	    this list of conditions and the following disclaimer in the documentation 
	    and/or other materials provided with the distribution. 

	THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
	FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
	THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
